/*
 * Copyright (C) 2013, 2014 Giorgio Vazzana
 *
 * This file is part of Seren.
 *
 * Seren is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Seren is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "nc.h"
#include "common.h"
#include "msgbook.h"

#define MODULE "ncurses"

struct NContext nctx;

static int create_windows()
{
	/* create title window */
	if ((nctx.win_title = newwin(1, nctx.cols, 0, 0)) == NULL)
		goto create_windows_fail;

	/* create message window */
	if ((nctx.win_message = newwin(nctx.rows - 4, nctx.cols - 17, 2, 0)) == NULL)
		goto create_windows_fail;
	scrollok(nctx.win_message, TRUE);

	/* create nodelist window */
	if ((nctx.win_nodelist = newwin(nctx.rows - 4, 16, 2, nctx.cols - 16)) == NULL)
		goto create_windows_fail;

	/* create status window */
	if ((nctx.win_status = newwin(1, nctx.cols, nctx.rows - 2, 0)) == NULL)
		goto create_windows_fail;

	/* create input window */
	if ((nctx.win_input = newwin(1, nctx.cols, nctx.rows - 1, 0)) == NULL)
		goto create_windows_fail;
	keypad(nctx.win_input, TRUE);

	return 0;

create_windows_fail:
	msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "Could not create ncurses window");
	return -1;
}

int nc_init(const char *version, const char *nick, int theme)
{
	initscr();
	if (has_colors() == FALSE) {
		endwin();
		msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "Your terminal does not support colors");
		return -1;
	}
	start_color();
	if (theme <=0 || theme > 2) { /* classic theme */
		/* main windows / messages colors */
		assume_default_colors(COLOR_WHITE, COLOR_BLUE);
		init_pair(1, COLOR_GREEN,  COLOR_BLACK);
		init_pair(2, COLOR_YELLOW, COLOR_BLACK);
		init_pair(3, COLOR_RED,    COLOR_BLACK);
		init_pair(4, COLOR_WHITE,  COLOR_BLUE);
		/* bars colors */
		init_pair(10, COLOR_BLACK,  COLOR_WHITE);
		init_pair(11, COLOR_GREEN,  COLOR_WHITE);
		init_pair(12, COLOR_YELLOW, COLOR_WHITE);
		init_pair(13, COLOR_RED,    COLOR_WHITE);
	} else if (theme == 1) { /* dark theme */
		assume_default_colors(COLOR_WHITE, COLOR_BLACK);
		init_pair(1, COLOR_GREEN,  COLOR_BLACK);
		init_pair(2, COLOR_YELLOW, COLOR_BLACK);
		init_pair(3, COLOR_RED,    COLOR_BLACK);
		init_pair(4, COLOR_CYAN,   COLOR_BLACK);

		init_pair(10, COLOR_BLACK,  COLOR_WHITE);
		init_pair(11, COLOR_GREEN,  COLOR_WHITE);
		init_pair(12, COLOR_YELLOW, COLOR_WHITE);
		init_pair(13, COLOR_RED,    COLOR_WHITE);
	} else if (theme == 2) { /* clear theme */
		assume_default_colors(COLOR_BLACK, COLOR_WHITE);
		init_pair(1, COLOR_GREEN,  COLOR_WHITE);
		init_pair(2, COLOR_YELLOW, COLOR_WHITE);
		init_pair(3, COLOR_RED,    COLOR_WHITE);
		init_pair(4, COLOR_CYAN,   COLOR_WHITE);

		init_pair(10, COLOR_WHITE,  COLOR_CYAN);
		init_pair(11, COLOR_WHITE,  COLOR_CYAN);
		init_pair(12, COLOR_YELLOW, COLOR_CYAN);
		init_pair(13, COLOR_RED,    COLOR_CYAN);
	}
	cbreak();
	noecho();
	getmaxyx(stdscr, nctx.rows, nctx.cols);

	if (create_windows() == -1) {
		endwin();
		return -1;
	}

	nc_print_bars(version);
	nc_clear_input_line_and_print_prompt(nick, 1);
	nctx.initialized = 1;

	return 0;
}

int nc_resize(const char *version, const char *nick)
{
	/* delete windows */
	delwin(nctx.win_title);
	delwin(nctx.win_message);
	delwin(nctx.win_status);
	delwin(nctx.win_input);

	/* reset ncurses and get new dimensions */
	endwin();
	refresh();
	getmaxyx(stdscr, nctx.rows, nctx.cols);
	clear();
	refresh();

	/* recreate windows */
	if (create_windows() == -1)
		return -1;

	nc_print_bars(version);
	snprintf(msgbuf, MBS, "Window resized to %dx%d", nctx.cols, nctx.rows);
	msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
	nc_clear_input_line_and_print_prompt(nick, 1);

	return 0;
}

void nc_close(void)
{
	/* shut down ncurses */
	if (nctx.initialized) {
		endwin();
		nctx.initialized = 0;
		mb0.print_cb = msgbook_print_cb_stderr;
		msgbook_flush(&mb0);
	}
}

void nc_print_bars(const char *version)
{
	/* title bar */
	mvwprintw(nctx.win_title, 0, 0, " Seren, ver. %s", version);
	mvwchgat(nctx.win_title, 0, 0, -1, A_NORMAL, 10, NULL);
	wnoutrefresh(nctx.win_title);

	/* status bar */
	mvwaddch(nctx.win_status, 0, 0, ' ');
	mvwchgat(nctx.win_status, 0, 0, -1, A_NORMAL, 10, NULL);
	wnoutrefresh(nctx.win_status);

	doupdate();
}

void nc_clear_input_line_and_print_prompt(const char *nick, int delete_buffer)
{
	if (delete_buffer) {
		nctx.linepos  = 0;
		nctx.wlinepos = 0;
	}
	mvwprintw(nctx.win_input, 0, 0, "%s> ", nick);
	wclrtoeol(nctx.win_input);
	wrefresh(nctx.win_input);
}

int nc_scroll_input_text(const char *nick, size_t right_space)
{
	int available_space;

	available_space = nctx.cols - ((int)strlen(nick) + 2) - /* leave a space on the right */ 1;
	if (available_space > 0 && nctx.wlinepos >= (size_t)available_space) {
		size_t i, j;

		nc_clear_input_line_and_print_prompt(nick, 0);
		for (i = nctx.wlinepos - (size_t)available_space + right_space; i < nctx.wlinepos; i++) {
			for (j = 0; j < utf8_bytes_in_sequence(nctx.wlinebuf[i][0]); j++)
				waddch(nctx.win_input, nctx.wlinebuf[i][j]);
		}
		return 1;
	}
	return 0;
}

void nc_status(int micmute, int record, int mode, float peak_percent, double dBSPL,
               const char *timer_str, const char *stat_str)
{
	int peakcolor, n, h;
	const char *mode_name[] = {"CLR", "SEC", "PSK"};

	if (micmute)
		peakcolor = 13;
	else
		peakcolor = (peak_percent <= 50.0f) ? 11 : ((peak_percent <= 80.0f) ? 12 : 13);

	wmove(nctx.win_status, 0, 0);
	wattron(nctx.win_status, COLOR_PAIR(10));
	wprintw(nctx.win_status, "\r%s[", timer_str);
	mode = mode <= 0 ? 0 : (mode > 2 ? 2 : mode);
	wattron(nctx.win_status, COLOR_PAIR(11+mode));
	waddstr(nctx.win_status, mode_name[mode]);
	wattron(nctx.win_status, COLOR_PAIR(10));
	waddch(nctx.win_status, '|');
	wattron(nctx.win_status, COLOR_PAIR(13));
	waddstr(nctx.win_status, record ? "REC" : "   ");
	wattron(nctx.win_status, COLOR_PAIR(10));
	waddstr(nctx.win_status, "]  mic[");
	n = micmute ? -90 : (int)lrint(dBSPL);
	n = (n < -50) ? 0 : (n+50); /* min -50dB */
	n = (n + 2) / 3;            /* 3dB step  */
	/*wprintw(nctx.win_status, "%+05.1f|", dBSPL);*/
	for (h = 0; h < 14; h++)
		waddch(nctx.win_status, h < n ? '#' : ' ');
	waddch(nctx.win_status, '|');
	wattron(nctx.win_status, COLOR_PAIR(peakcolor));
	if (micmute)
		wprintw(nctx.win_status, "MUTE");
	else
		wprintw(nctx.win_status, "%3.0f%%", peak_percent);
	wattron(nctx.win_status, COLOR_PAIR(10));
	wprintw(nctx.win_status, "]  %s", stat_str);
	wattroff(nctx.win_status, COLOR_PAIR(10));
	wchgat(nctx.win_status, -1, A_NORMAL, 10, NULL);

	wnoutrefresh(nctx.win_status);
	wnoutrefresh(nctx.win_input);
	doupdate();
}

void nc_nodelist(unsigned int n, const char *nicks[], int mute[], double dBSPL[],
                 unsigned int pl10k[], unsigned int tm_pl10k[])
{
	unsigned int i;

	werase(nctx.win_nodelist);
	mvwaddstr(nctx.win_nodelist, 0, 0, "      Nodes\n\n");
	for (i = 0; i < n; i++) {
		int n, h;
		if (nicks[i]) {
			if (i == 0) {
				wprintw(nctx.win_nodelist, " * %s\n ", nicks[0]);
			} else {
				int health;
				unsigned int pl10k_max;

				pl10k_max = pl10k[i] >= tm_pl10k[i] ? pl10k[i] : tm_pl10k[i];
				health = pl10k_max <= 5000U ? 0 : (pl10k_max <= 50000U ? 1 : 2);

				wattron(nctx.win_nodelist, COLOR_PAIR(1+health));
				wprintw(nctx.win_nodelist, " %d %s\n ", i-1, nicks[i][0] == '\0' ? "?" : nicks[i]);
				wattroff(nctx.win_nodelist, COLOR_PAIR(1+health));
			}

			waddch(nctx.win_nodelist, ACS_LLCORNER);
			n = mute[i] ? -90 : (int)lrint(dBSPL[i]);
			n = (n < -50) ? 0 : (n+50); /* min -50dB */
			n = (n + 2) / 3;            /* 3dB step  */
			/*wprintw(nctx.win_nodelist, "%+05.1f|", dBSPL[i]);*/
			for (h = 0; h < 14; h++)
				waddch(nctx.win_nodelist, h < n ? '#' : ' ');
		}
	}
	mvwvline(nctx.win_nodelist, 0, 0, ACS_VLINE, nctx.rows - 4 - 1);
	mvwaddch(nctx.win_nodelist, 1, 0, ACS_LTEE);
	mvwhline(nctx.win_nodelist, 1, 1, ACS_HLINE, 16);
	wnoutrefresh(nctx.win_nodelist);
	wnoutrefresh(nctx.win_input);
	doupdate();
}


void nc_print_cb(time_t ts, unsigned int type, const char *module, const char *msg)
{
	if (type == MB_TYPE_DEBUG || type == MB_TYPE_VERBOSE) {
		wattron(nctx.win_message, COLOR_PAIR(1));
		wprintw(nctx.win_message, "[%s][%s] %s\n", ts2time(ts), module, msg);
		wattroff(nctx.win_message, COLOR_PAIR(1));
	} else if (type == MB_TYPE_WARNING) {
		wattron(nctx.win_message, COLOR_PAIR(2));
		wprintw(nctx.win_message, "[%s][%s] %s\n", ts2time(ts), module, msg);
		wattroff(nctx.win_message, COLOR_PAIR(2));
	} else if (type == MB_TYPE_ERROR) {
		wattron(nctx.win_message, COLOR_PAIR(3));
		wprintw(nctx.win_message, "[%s][%s] %s\n", ts2time(ts), module, msg);
		wattroff(nctx.win_message, COLOR_PAIR(3));
	} else if (type == MB_TYPE_GUI) {
		wattron(nctx.win_message, COLOR_PAIR(4));
		wprintw(nctx.win_message, "[%s][%s] %s\n", ts2time(ts), module, msg);
		wattroff(nctx.win_message, COLOR_PAIR(4));
	} else if (type == MB_TYPE_CHAT) {
		wattron(nctx.win_message, A_BOLD);
		wprintw(nctx.win_message, "[%s] %s> %s\n", ts2time(ts), module, msg);
		wattroff(nctx.win_message, A_BOLD);
	} else {
		wprintw(nctx.win_message, "[%s][%s] %s\n", ts2time(ts), module, msg);
	}

	wnoutrefresh(nctx.win_message);
	wnoutrefresh(nctx.win_input);
	doupdate();
}

#undef MODULE
